package App::Logger;

# debug:   write if /debug/
# info:    write
# warning: write
# error:   write
# fatal:   write

use strict;
use warnings;

use Carp;
use Data::Dumper;

use Sys::Hostname;

sub new {
  my ($class, %args) = @_;

  my $self = {};
  # mandatory
  $self->{_file} = $args{file} or croak "no file given";
  # optional
  $self->{_debug} = $args{debug} ? 1 : 0;
  $self->{_verbosity} = $args{verbosity};

  bless($self, $class);

  $self->_init();

  return $self;
}

sub _init {
  my ($self) = @_;
  croak "_init is private" if caller() ne __PACKAGE__;

  # settings
  $self->{_user} = $ENV{USER} || '<undef>';
  $self->{_hostname} = $ENV{HOSTNAME} || hostname() || '<undef>';

  # levels
  my $levels = {
    debug => { on_debug_only => 1 },
    info => {},
    warning => {},
    error => {},
    fatal => {},
  };

  # create subs on the fly
  LEVEL:
  for my $level ( keys %$levels ) {

    # define sub /level/
    my $sub = sub {
      my ($self, $msg, $verbosity) = @_;
      $verbosity ||= 0;

      # act according to /verbosity/
      if ( $self->{_verbosity} >= $verbosity ) {

        # some /level/ should write /on_debug_only/ if /_debug/
        if ( $levels->{ $level }->{on_debug_only} ) {
          $self->_write($msg, $level) if $self->{_debug};
        }
        else {
          $self->_write($msg, $level);
        }

        # put extra /level/ stuff here ..

      }

      return;
    };

    # create sub /level/
    do {
      no strict 'refs';
      no warnings 'redefine';
      *{$level} = $sub;
    };

  } # LEVEL

  return;
}

sub _write {
  my ($self, $msg, $level) = @_;
  croak "_write is private" if caller() ne __PACKAGE__;

  $msg ||= '';
  chomp $msg;

  my $scalar_localtime = scalar localtime;
  my $pid = $$;

  my $message = $scalar_localtime                     # timestamp
              . ' '                                   # whitespace
              . "[$self->{_hostname}]"                # hostname
              . ' '                                   # whitespace
              . "[$pid] [$level]"                     # info
              . ( $self->{_debug} ? ' {debug}' : '' ) # debug?
              . ': '                                  # separator, whitespace
              . $msg                                  # msg
              . "\n"                                  # newline
              ;

  open my $fh, '>>', $self->{_file} or croak "no open at file |$self->{_file}|, |$!|";
    print $fh $message or croak "no print at file |$self->{_file}|, |$!|";
  close $fh or carp "no close at file |$self->{_file}|, |$!|";

  return;
}

1;

__END__

