package App::Type::OAIHarvest_LOM;

# workout job type OAIHarvest_LOM

use strict;
use warnings;

use Carp;
use Data::Dumper;

use FindBin;
use lib "$FindBin::Bin/../../";
use App::Logger;
use JSON::XS;
use Net::AMQP::RabbitMQ;
use Time::HiRes qw/sleep/;
use Net::OAIPMH::Harvester;

sub new {
  my ($class, %args) = @_;

  my $self = {};
  # mandatory
  $self->{_cfg} = $args{cfg} || croak "no cfg given";

  bless($self, $class);

  $self->_init();

  return $self;
}

sub _init {
  my ($self) = @_;
  croak "_init is private" if caller() ne __PACKAGE__;

  # new logger
  $self->{_logger} = App::Logger->new(
    file => $FindBin::Bin.'/'.$self->{_cfg}->{log}->{path}.'/'.$self->{_cfg}->{log}->{filename},
    debug => defined $self->{_cfg}->{debug} ? $self->{_cfg}->{debug} : 1,
    verbosity => defined $self->{_cfg}->{verbosity} ? $self->{_cfg}->{verbosity} : 2,
  );

  return;
}

sub run {
  my ($self) = @_;

  ## settings
  # shortcut
  my $logger = $self->{_logger};

  $logger->info('OAIHarvest_LOM start', 1);

  # this method can die/croak/confess

  eval {

    $logger->debug('_cfg ' . Dumper( $self->{_cfg} ) );

    $logger->info('OAIHarvest_LOM start', 1);

    # OAI-PMH source
    $logger->info('harvester Net::OAIPMH::Harvester->new', 2);
    my $uri = $self->{_cfg}->{uri}->{scheme} . ':' . $self->{_cfg}->{uri}->{authority}
              . $self->{_cfg}->{uri}->{path};
    my $harvester = Net::OAIPMH::Harvester->new( baseURL => $uri );

    # MessageQueue target
    $logger->info('mq Net::AMQP::RabbitMQ->new', 2);
    my $mq = Net::AMQP::RabbitMQ->new();
    $logger->info('mq connect', 2);
    $mq->connect(
      # hostname
      $self->{_cfg}->{mq}->{host},
      # options
      {
        port => $self->{_cfg}->{mq}->{port},
        user => $self->{_cfg}->{mq}->{user},
        password => _slurp( $self->{_cfg}->{mq}->{pass} ),
        vhost => $self->{_cfg}->{mq}->{vhost},
        timeout => 5, # <42> seconds, default undef
      },
    );
    $logger->info('mq channel_open', 2);
    $mq->channel_open(1);

    $logger->info('token start', 2);

    my $c = 0; # count objects
    my $mq_publish_expiration = $self->{_cfg}->{sleep}->{post} * 1000; # milliseconds

    ##### collect data via oai_pmh starting here:

    $logger->info("request metadataPrefix |$self->{_cfg}->{uri}->{parameters}->{prefix}|", 2);
    my %opts = ( 'metadataPrefix'	=> $self->{_cfg}->{uri}->{parameters}->{prefix} );
    # $opts{ 'from' } = $self->{_cfg}->{uri}->{parameters}->{from}; # expacts a start date not used
    # $opts{ 'until' } = $self->{_cfg}->{uri}->{parameters}->{from}; # expacts a end date not used
    if (  $self->{_cfg}->{uri}->{parameters}->{set} ne 'noSetHierarchy' ) { 
    	$opts{ 'set' } = $self->{_cfg}->{uri}->{parameters}->{set};
    }
    $opts{ 'metadataHandler' } = $self->{_cfg}->{uri}->{parameters}->{handler};
    $logger->info("request set |$self->{_cfg}->{uri}->{parameters}->{set}|", 2);

    # request the server at
    $logger->info("request oai uri |$uri|", 2);
    $logger->info("request oai opts |". Dumper( %opts ) ."|", 2);
      
    my $records = $harvester->listRecords( %opts );
    my $finished = 0;
    my $completeListSize = 0;

    TOKEN:
    # while
    while ( ! $finished ) { 

      if ( $records->errorCode() ) { 

        $logger->info('records errorCode', 2);
        if ( $records->errorCode() eq 'xmlParseError' ) { 
          $logger->error('xml Parser Error ' . Dumper( $records->file() ) );
        }
        $logger->error("no valid response from oai uri: $uri | code: ".$records->errorCode(). " | reason: ".$records->errorString()." |");
        $logger->info('last token', 2);
        last TOKEN;
      }

      while ( my $record = $records->next() ) {
        $logger->info('header identifier: ' . $record->header()->identifier(), 1);

        $logger->debug('object ' . Dumper( $record->metadata() ) );

        ###### get the objects and push them into the message queue
        OBJECTS:
        my $r = $record->metadata();

        my $data = encode_json( $r->lom() );
        $logger->debug('json data' . $data );

        $logger->info('mq publish', 2);
        $mq->publish(
          # channel
          1,
          # routing_key
          $self->{_cfg}->{mq}->{routing_key},
          # body
          $data,
          # options
          {
            exchange => '', # default 'amq.direct'
          },
          # props
          {
            delivery_mode => 2,
            expiration => $mq_publish_expiration, # milliseconds
          },
        );
        
        # $main::term
        if ( $main::term ) {
          last TOKEN;
        }         

        $c++;
      }

      my $resumptionToken = $records->resumptionToken();
      if ( $resumptionToken ) {        
        $logger->info("sleep before next token |".$self->{_cfg}->{sleep}->{token}."|", 2);
        sleep $self->{_cfg}->{sleep}->{token};
        $logger->info("using resumption token:" . $resumptionToken->token(), 2);
        $completeListSize = defined $resumptionToken->completeListSize() ? $resumptionToken->completeListSize() : -1;
        my $cursor = $resumptionToken->cursor();
        $logger->info("completeListSize:" . $completeListSize . "cursor:" . $cursor, 2);
        $opts{ 'resumptionToken' } = $resumptionToken->token();
        $logger->info("request oai uri |$uri|", 2);
        $logger->info("request oai opts |". Dumper( %opts ) ."|", 2);
        $records = $harvester->listRecords( 
          resumptionToken => $resumptionToken->token(),
          metadataHandler => $self->{_cfg}->{uri}->{parameters}->{handler}
        );
      } else { 
        $finished = 1;
      }
    }



    $logger->warning('SIG via main::term received', 2) if $main::term;

    $logger->info('token done', 2);

    $logger->info("c |$c| vs completeListSize |$completeListSize|", 1);

    $logger->info('mq disconnect', 2);
    $mq->disconnect();

    $logger->info('OAIHarvest_LOM workout done', 1);

  }; # eval
  if ( $@ ) {
    my $syntax_error_message = $@;
    chomp $syntax_error_message;
    $logger->fatal("OAIHarvest_LOM died |$syntax_error_message|");
    croak $syntax_error_message;
  }

  $logger->info('OAIHarvest_LOM done', 1);

  return;
}

sub _slurp {
  my ($file) = @_;

  my $slurp;
  open my $fh, '<', $file or croak "no open at file |$file|, |$!|";
    do {
      local $/;
      $slurp = <$fh>;
    };
  close $fh or carp "no close at file |$file|, |$!|";
  chomp $slurp;

  return $slurp;
}

1;

__END__

