package App;

# workout queues

use strict;
use warnings;

use Carp;
use Data::Dumper;

use YAML::Syck;
use FindBin;
use lib "$FindBin::Bin";
use App::Logger;
use App::Type::Openlib;
use App::Type::Phaidra;
use App::Type::OAIIngest_DC;
use App::Type::OAIIngest_LOM_edusharing; # for uibk and uni-graz
use App::Type::OAIIngest_LOM_invenio; # for tugraz
use App::Type::dummy;
use LWP::Simple;
use List::MoreUtils qw/none/;
use Net::AMQP::RabbitMQ;
use Time::HiRes qw/sleep/;

sub new {
  my ($class, %args) = @_;

  my $self = {};
  # mandatory
  $self->{_ctrl} = $args{ctrl} or croak "no ctrl given";
  $self->{_queues} = $args{queues} or croak "no queues given";

  bless($self, $class);

  $self->_init();
  $self->_sanity();

  return $self;
}

sub _init {
  my ($self) = @_;
  croak "_init is private" if caller() ne __PACKAGE__;

  # new logger
  $self->{_logger} = App::Logger->new(
    file => $FindBin::Bin.'/'.$self->{_ctrl}->{log}->{path}.'/'.$self->{_ctrl}->{log}->{filename},
    debug => defined $self->{_ctrl}->{debug} ? $self->{_ctrl}->{debug} : 1,
    verbosity => defined $self->{_ctrl}->{verbosity} ? $self->{_ctrl}->{verbosity} : 2,
  );

  return;
}

sub _sanity {
  my ($self) = @_;
  croak "_sanity is private" if caller() ne __PACKAGE__;

  ## settings
  # shortcut
  my $logger = $self->{_logger};

  # config
  my $sanity = YAML::Syck::LoadFile("$FindBin::Bin/etc/queues/sanity.yml");

  $logger->info('sanity check start', 2);
  $logger->debug('_ctrl ' . Dumper( $self->{_ctrl} ) );
  $logger->debug('_queues ' . Dumper( $self->{_queues} ) );

  QUEUE:
  for my $queue ( keys %{$self->{_queues}} ) {
    croak "sanity check died, unknown queue type |$self->{_queues}->{ $queue }->{type}|"
      if none { $_ eq $self->{_queues}->{ $queue }->{type} } @{$sanity->{types}};
  }

  $logger->info('sanity check done', 2);

  return;
}

# return: SCALAR; 0:no error, 1:error
sub run {
  my ($self) = @_;

  ## settings
  # shortcut
  my $logger = $self->{_logger};

  $logger->info('App start');

  # to monitor failure of eval block
  # this method shall not die/croak
  my $error;

  eval {

    $logger->info('App workout start', 1);

    $logger->info("sleep pre |$self->{_ctrl}->{sleep}->{pre}|", 1);
    sleep $self->{_ctrl}->{sleep}->{pre};

    # MessageQueue source
    $logger->info('mq Net::AMQP::RabbitMQ->new', 2);
    my $mq = Net::AMQP::RabbitMQ->new();
    $logger->info('mq connect', 2);
    $mq->connect(
      # hostname
      $self->{_ctrl}->{mq}->{host},
      # options
      {
        port => $self->{_ctrl}->{mq}->{port},
        user => $self->{_ctrl}->{mq}->{user},
        password => _slurp( $self->{_ctrl}->{mq}->{pass} ),
        vhost => $self->{_ctrl}->{mq}->{vhost},
        timeout => 5, # <42> seconds, default undef
      },
    );
    $logger->info('mq channel_open', 2);
    $mq->channel_open(1);

    $logger->info("mq basic_qos prefetch_count |$self->{_ctrl}->{mq}->{prefetch_count}|", 2);
    $mq->basic_qos(
      # channel
      1,
      # options
      {
        prefetch_count => $self->{_ctrl}->{mq}->{prefetch_count},
      },
    );

    # filter out queuename 'dummy'
    $logger->info('queues', 2);
    my @queues;
    QUEUE:
    for my $queue ( keys %{$self->{_queues}} ) {
      next QUEUE if $self->{_queues}->{ $queue }->{type} eq 'dummy';
      push @queues, $queue;
    }
    $logger->debug('queues ' . Dumper( \@queues ) );

    $logger->info('mq consume', 2);
    QUEUENAME:
    for my $queuename ( @queues ) {
      $logger->info("mq consume queuename |$queuename|", 2);
      my $consumer_tag = $mq->consume(
        # channel
        1,
        # queuename
        $queuename,
        # options
        {
          no_ack => 0, # default 1
        },
      );
      $logger->info("consumer_tag |$consumer_tag|", 2);
    }

    # init obj class 'App::Type::/type/'
    #   store objs->{ /queue/ } = obj
    #   /queue/ == msg /routing_key/
    #   hence, retreive obj via objs->{ /routing_key/ }->run()
    $logger->info('objs', 2);
    my $objs;
    QUEUE:
    for my $queue ( keys %{$self->{_queues}} ) {
      $logger->info("queue |$queue|", 2);

      my $type = $self->{_queues}->{ $queue }->{type};
      $logger->info("type |$type|", 2);

      my $class = 'App::Type::'.$type;
      $logger->info("class |$class|", 2);

      $logger->info('obj App::Type::'.$type.'->new', 2);
      my $obj = $class->new(
        cfg => $self->{_queues}->{ $queue }->{cfg},
      );
      $objs->{ $queue } = $obj;
    }

    # UserAgent target
    $logger->info('ua LWP::UserAgent->new', 2);
    my $ua = LWP::UserAgent->new();
    $ua->timeout(10);

    # objects max
    if ( $self->{_ctrl}->{objects}->{max} ) {
      $logger->info("objects max |$self->{_ctrl}->{objects}->{max}|", 2 );
    }

    $logger->info('recv start', 1);

    my $c = 0; # count objects
    $logger->info('receiving ..', 2);
    RECV:
    while ( !$main::term && ( my $msg = $mq->recv() ) ) {

      $logger->info('msg received', 2);
      $logger->debug('msg ' . Dumper( $msg ) );

      $logger->info('msg workout start', 2);

      $logger->info("msg routing_key |$msg->{routing_key}|", 2);
      $logger->info("msg delivery_tag |$msg->{delivery_tag}|", 2);

      $logger->info("msg routing_key |$msg->{routing_key}| start", 2);
      $objs->{ $msg->{routing_key} }->run(
        ua => $ua,
        msg => $msg,
      );
      $logger->info("msg routing_key |$msg->{routing_key}| done", 2);

      $logger->info("mq ack delivery_tag |$msg->{delivery_tag}|", 2);
      $mq->ack(
        # channel
        1,
        # delivery_tag
        $msg->{delivery_tag},
      );

      $logger->info('msg workout done', 2);

      $c++;
      $logger->info("c |$c|", 2);

      # objects max
      if ( $self->{_ctrl}->{objects}->{max} ) {
        if ( $c >= $self->{_ctrl}->{objects}->{max} ) {
          $logger->info('c >= objects max', 2);
          $logger->info('last recv', 2);
          last RECV;
        }
      }

      $logger->info("sleep batch |$self->{_ctrl}->{sleep}->{batch}|", 2);
      sleep $self->{_ctrl}->{sleep}->{batch};

      $logger->info('receiving ..', 2);

    } # while ( !$main::term && ( my $msg = $mq->recv() ) )

    $logger->warning('SIG via main::term received', 2) if $main::term;

    $logger->info('recv done', 1);

    $logger->info("c |$c|", 1);

    $logger->info('mq disconnect', 2);
    $mq->disconnect();

    $logger->info('App workout done', 1);

  }; # eval
  if ( $@ ) {
    my $syntax_error_message = $@;
    chomp $syntax_error_message;
    $error++;
    $logger->error("App died |$syntax_error_message|");
  }

  $logger->info('App done');

  return $error ? 1 : 0;
}

sub _slurp {
  my ($file) = @_;

  my $slurp;
  open my $fh, '<', $file or croak "no open at file |$file|, |$!|";
    do {
      local $/;
      $slurp = <$fh>;
    };
  close $fh or carp "no close at file |$file|, |$!|";
  chomp $slurp;

  return $slurp;
}

1;

__END__

