package App::Type::OAIIngest_LOM_invenio;

# workout job type OAIIngest_LOM

use strict;
use warnings;

use Carp;
use Data::Dumper;

use FindBin;
use lib "$FindBin::Bin/../../";
use Encode;
use App::Logger;
use App::Type::LOMIngest_invenio;
use App::Type::Vocabs;
use JSON::XS;
use LWP::Simple;
use Text::CSV_XS;
# use XML::Simple; Not Used
use utf8;

sub new {
  my ($class, %args) = @_;

  my $self = {};
  # mandatory
  $self->{_cfg} = $args{cfg} || croak "no cfg given";

  bless($self, $class);

  $self->_init();

  return $self;
}

sub _init {
  my ($self) = @_;
  croak "_init is private" if caller() ne __PACKAGE__;

  # new logger
  $self->{_logger} = App::Logger->new(
    file => $FindBin::Bin.'/'.$self->{_cfg}->{log}->{path}.'/'.$self->{_cfg}->{log}->{filename},
    debug => defined $self->{_cfg}->{debug} ? $self->{_cfg}->{debug} : 1,
    verbosity => defined $self->{_cfg}->{verbosity} ? $self->{_cfg}->{verbosity} : 2,
  );

  $self->{_logger}->debug('_cfg ' . Dumper( $self->{_cfg} ) );

  $self->{_uri} = defined $self->{_cfg}->{uri}  ? $self->{_cfg}->{uri}->{scheme} . ':' . $self->{_cfg}->{uri}->{authority} . $self->{_cfg}->{uri}->{path} : 'http://index:9200/invenio';
  $self->{_age} = defined $self->{_cfg}->{ingest}  ? $self->{_cfg}->{ingest}->{age} : '12h';
  return;
}

sub run {
  my ($self, %args) = @_;

  # args
  $self->{_ua} = $args{ua} || croak "no ua given";
  $self->{_msg} = $args{msg} || croak "no msg given";

  ## settings
  # shortcut
  my $logger = $self->{_logger};

  $logger->info('OAIIngest_LOM start', 1);

  # this method can die/croak/confess

  eval {

    $logger->debug('_msg ' . Dumper( $self->{_msg} ), 2);

    $logger->info('OAIIngest_LOM workout start', 1);

    $logger->info('prepare metadata', 2);

    # apparently Net::AMQP::RabbitMQ->recv() returns a Perl data structure
    # hence, we have to encode it to UTF-8 before decode_json
    #
    # https://perldoc.perl.org/Encode.html
    #   "Encodes [..] from Perl's internal form into ENCODING"
    # Perl -> UTF-8
    my $body = encode( 'UTF-8', $self->{_msg}->{body} );
    $logger->debug('body ' . Dumper( $body ), 2);
    $logger->info('OAIIngest_LOM decode body', 2);

    # https://perldoc.perl.org/Encode.html
    #   "Encodes [..] from Perl's internal form into ENCODING"
    # Perl -> UTF-8
    my $item = decode_json( $body );
    # body has LOM as json
    
    my $id = getId($item, $self->{_msg}->{routing_key}) || ""; # 'repository.tugraz.at'
    if ($id eq "") {
      $logger->info('processing id[] skipped, reason unknown id!', 1);
    } else {
      $logger->info('processing id[' . $id .']', 1);
      
      # custom filter
      $logger->info('custom filter', 2);

      #$item->{oea_title} = getTitle($item);
      $item->{oea_title} = getTitleMostRelated($item);
      $item->{oea_title_ml} = getLangTitle($item);
      #$item->{oea_abstract} = getAbstract($item); # do not use this anymore
      #$item->{oea_abstract} = getAbstractMostRelated($item);
      my $abs = getAbstractMostRelated($item);
      if (length($abs) gt 0) {
        $item->{oea_abstract} = $abs;
      } else {
        $item->{oea_abstract} = "Keine Beschreibung vorhanden!";
      }
      $item->{oea_abstract_ml} = getLangAbstract($item);
      my $aut = getAuthors($item);
      if (scalar @$aut gt 0) {
        $item->{oea_authors} = getAuthors($item);
      } else {
        $item->{oea_authors} = getPublishers($item);
      }
      $item->{oea_object_direct_link} = getDOILink($item) || getDirectLink($item);
      $item->{oea_thumbnail_url} = getThumbnailUrl($item) || "";
      $item->{oea_classification_00} = getFormat($item);
      $item->{oea_classification_05} = getTechnicalFormat($item);
      $item->{oea_ingest} = ingestTimeStamp();

      my $txt_de = '';
      my $txt_en = '';
      my $tag = '';
      my $obj = {};

      my $vocabsOefos = App::Type::Vocabs->new( vocabs => 'vocabs.oefos2012');
      my $list = getOefosList($item);
      my $subCatList = onlyOefosSubCategories($list);
      my $disciplines;
      for my $i ( @$subCatList ) {
        if ($i eq "") {
          $obj = {
            id => '999999',
            name_de => "Sonstiges",
            name_en => "Other",
            tag => "OEAUNDEF"
          };
        }
        else {
          $txt_de = $vocabsOefos->getVocab($i, 'name_de');
          $txt_en = $vocabsOefos->getVocab($i, 'name_en');
          $tag = $vocabsOefos->getVocab($i, 'tag');
          $obj = {
            id => $i,
            name_de => $txt_de,
            name_en => $txt_en,
            tag => $tag
          };
        }
        push (@$disciplines, $obj);
      }
      $item->{oea_classification_01} = \@$disciplines;

      my $vocabsSpdx = App::Type::Vocabs->new( vocabs => 'vocabs.spdx');
      my $lic = getLicense($item);
      my $spdx = $vocabsSpdx->getVocab($lic, 'name_en');
      if ( $lic eq "" ) {
        $item->{oea_valid} = 'false';
        $item->{oea_unvalid_remark} = 'empty license string';
      } elsif ( $spdx eq "" ) {
        $item->{oea_valid} = 'false';
        $item->{oea_unvalid_remark} = 'no valid license: '.$lic;
      } else {
        $item->{oea_classification_02} = $lic;
        $item->{oea_valid} = 'true';
      }

      $item->{oea_classification_03} = getCreated($item); # $item->{created};

      my $vocabsLearningRT = App::Type::Vocabs->new( vocabs => 'vocabs.kim.hcrt');
      my $learningresourcetypeList = getLearningRTList($item);
      my $learningresourcetypes;
      for my $i ( @$learningresourcetypeList ) {
        if ($i eq "") {
          $obj = {
            id => $i,
            name_de => "Nicht bekannt",
            name_en => "Not defined",
            tag => "OEAUNDEF"
          };
        }
        else {
          $txt_de =  $vocabsLearningRT->getVocab($i, 'name_de');
          $txt_en =  $vocabsLearningRT->getVocab($i, 'name_en');
          $tag =  $vocabsLearningRT->getVocab($i, 'tag');
          # HCRT Hochschulcampus ResourcenTypen https://github.com/dini-ag-kim/hcrt
          $obj = {
            id => $i,
            name_de => $txt_de,
            name_en => $txt_en,
            tag => $tag
          };
        }
        push (@$learningresourcetypes, $obj);
      }
      $item->{oea_classification_04} = \@$learningresourcetypes;
      $item->{oea_classification_05} = getTechnicalFormat($item);

      my $vocabsLanguage = App::Type::Vocabs->new( vocabs => 'vocabs.iso639');
      my $languageList = getLanguageList($item);
      my $languages;
      my $language2digit;
      for my $i ( @$languageList ) {
        if ($i eq "") {
          $obj = {
            id => "999999",
            name_de => "Nicht bekannt",
            name_en => "Not defined",
            tag => "OEAUNDEF"
          };
        }
        else {
          $language2digit= substr($i, 0, 2);
          $txt_de =  $vocabsLanguage->getVocab($language2digit, 'name_de');
          $txt_en =  $vocabsLanguage->getVocab($language2digit, 'name_en');
          $tag =  $vocabsLanguage->getVocab($language2digit, 'tag');
          $obj = {
            id => $language2digit,
            name_de => $txt_de,
            name_en => $txt_en,
            tag => $tag
          };
        }
        push (@$languages, $obj);
      }
      $item->{oea_classification_06} = \@$languages;

      # custom delete
      foreach my $titles  ( $item->{general}->{title} ) {
        my $titles_type = ref($titles);
        my $j = 0;
        foreach my $i (@$titles) {
          my $i_type = ref($i);
          if ( $i_type ne "HASH" ) {
            $logger->info('delete general.title [' .$j. '] - reason:mapper_parsing_exception', 1);
            delete $item->{general}->{title}[$j];
          }
          $j++;
        }
      }

      for my $keywords  ( $item->{general}->{keywords} ) {
        my $keywords_type = ref($keywords);
        my $j = 0;
        foreach my $i (@$keywords) {
          if ( $keywords_type ne "HASH" ) {
            $logger->info('delete general.keywords [' .$j. '] - reason:mapper_parsing_exception', 1);
            delete $item->{general}->{keywords}[$j];
          }
          $j++;
        }
      }
    
      {
        # settings
        local $Data::Dumper::Indent = 0;
        $logger->debug('item ' . Dumper( $item ), 2);
      }

      my $uri = $self->{_uri} . "/_doc/$id";
      $logger->info("uri |$uri|", 2);

      # encode_json item as $data
      $logger->info('encode_json item', 2);
      my $data = encode_json( $item );

      {
        # settings
        local $Data::Dumper::Indent = 0;
        $logger->debug('data ' . Dumper( $data ), 2);
      }

      $logger->info('req HTTP::Request->new', 2);
      my $req = HTTP::Request->new( 'PUT', $uri );
      $req->header( 'Content-Type' => 'application/json' );
      $req->content( $data );

      $logger->info('ingest start', 2);
      my $res = $self->{_ua}->request( $req );
      if ( !$res->is_success() ) {
        $logger->fatal("no valid response from uri |$uri|, status_line |".$res->status_line()."|, content |".$res->content()."|");
        croak "no valid response from uri |$uri|, status_line |".$res->status_line()."|, content |".$res->content()."|";
      }
      $logger->info('res status_line |'.$res->status_line().'|', 2);
      $logger->info('res content |'.$res->content().'|', 2);
      $logger->info('ingest done', 2);
    }
    # remove old data
    my $uri_delete = $self->{_uri} . "/_delete_by_query";
    $logger->info("uri |$uri_delete|", 2);    
    my $req_delete = HTTP::Request->new( 'POST', $uri_delete );
    $req_delete->header( 'Content-Type' => 'application/json' );
    my $del_query = '{ "query": { "range": { "oea_ingest": { "lte": "now-'.$self->{_age}.'" } } } }';
    $req_delete->content( $del_query );

    $logger->info('delete start (lte '.$self->{_age}.')', 2);
    my $res_delete = $self->{_ua}->request( $req_delete );
    if ( !$res_delete->is_success() ) {
      $logger->fatal("no valid response from uri |$uri_delete|, status_line |".$res_delete->status_line()."|, content |".$res_delete->content()."|");
      # croak "no valid response from uri |$uri_delete|, status_line |".$res_delete->status_line()."|, content |".$res_delete->content()."|";
    }
    $logger->info('res status_line |'.$res_delete->status_line().'|', 2);
    $logger->info('res content |'.$res_delete->content().'|', 2);
    $logger->info('delete done', 2);    

    $logger->info('OAIIngest_LOM workout done', 1);
  }; # eval
  if ( $@ ) {
    my $syntax_error_message = $@;
    chomp $syntax_error_message;
    $logger->fatal("OAIIngest_LOM died |$syntax_error_message|");
    croak $syntax_error_message;
  }

  $logger->info('OAIIngest_LOM done', 1);

  return;
}

1;

__END__
